/*

RemindMe Application

*/

#include <Carbon/Carbon.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "date.hpp"
#include "app.hpp"

static WindowRef Window = 0;
static const ControlID DateControlID = {'REME', 3};
static const ControlID StaticControlID = {'REME', 6};
static const ControlID ViewControlID = {'REME', 7};
static const ControlID BeforeStaticID = {'REME', 8};
static const ControlID BeforeSliderID = {'REME', 9};
static const ControlID AfterStaticID = {'REME', 10};
static const ControlID AfterSliderID = {'REME', 11};
static HIViewRef TextView = 0;
static TXNObject TextObject = 0;

void UISetDate(RMDateType dt)
{
    LongDateRec ldr;
    ControlRef cr;
    Size sz;

    GetControlByID(Window, &DateControlID, &cr);
    GetControlData(cr, kControlEntireControl, kControlClockLongDateTag,
            sizeof(ldr), &ldr, &sz);

    ldr.ld.day = (short) dt.DayOfMonth;
    ldr.ld.month = (short) dt.Month;
    ldr.ld.year = (short) dt.Year;

    SetControlData(cr, kControlEntireControl, kControlClockLongDateTag,
            sizeof(ldr), &ldr);
    HIViewSetNeedsDisplay(cr, 1);
}

RMDateType UIGetDate()
{
    RMDateType dt;
    LongDateRec ldr;
    ControlRef cr;
    Size sz;

    GetControlByID(Window, &DateControlID, &cr);
    GetControlData(cr, kControlEntireControl, kControlClockLongDateTag,
            sizeof(ldr), &ldr, &sz);

    dt.DayOfMonth = ldr.ld.day;
    dt.Month = ldr.ld.month;
    dt.Year = ldr.ld.year;

    return(dt);
}

static const char * DayOfWeekNames[7] = {"Sunday", "Monday", "Tuesday",
    "Wednesday", "Thursday", "Friday", "Saturday"};

static const char * MonthNames[12] = {"January", "February", "March",
    "April", "May", "June", "July", "August", "September", "October",
    "November", "December"};

void UIUpdateDate(RMDateType dt)
{
    char buf[128];
    ControlRef cr;
    CFStringRef sr;

    sprintf(buf, "%s %d %s %d", DayOfWeekNames[DayOfWeek(dt.DayOfMonth,
            dt.Month, dt.Year) - 1], dt.DayOfMonth, MonthNames[dt.Month - 1],
            dt.Year);

    GetControlByID(Window, &StaticControlID, &cr);

    sr = CFStringCreateWithBytes(0, (const UInt8 *) buf, strlen(buf),
            kCFStringEncodingMacRoman, 0);
    SetControlData(cr, kControlEntireControl,
            kControlStaticTextCFStringTag, sizeof(CFStringRef), &sr);
    CFRelease(sr);
    HIViewSetNeedsDisplay(cr, 1);
}

int UIGetBefore()
{
    ControlRef cr;

    GetControlByID(Window, &BeforeSliderID, &cr);
    return(GetControl32BitValue(cr));
}

int UIGetAfter()
{
    ControlRef cr;

    GetControlByID(Window, &AfterSliderID, &cr);
    return(GetControl32BitValue(cr));
}

void UIClearText()
{
    TXNSetData(TextObject, kTXNTextData, 0, 0, kTXNStartOffset,
            kTXNEndOffset);
}

void UIAppendText(char * s)
{
    TXNSetData(TextObject, kTXNTextData, s, strlen(s),
            kTXNEndOffset, kTXNEndOffset);
}

static const char * ShortMonthNames[12] = {"jan", "feb", "mar",
    "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"};

static char RunRMEPath[PATH_MAX];

void RunErrorMessage()
{
    UIClearText();
    UIAppendText("Error executing runrme\n");
    UIAppendText(RunRMEPath);
}

void UIRunRME(RMDateType dt, int b, int a)
{
    int rwp[2];
    int wrp[2];
    pid_t pid;

    if (pipe(rwp) != 0)
    {
        RunErrorMessage();
        return;
    }

    if (pipe(wrp) != 0)
    {
        close(rwp[0]);
        close(rwp[1]);
        RunErrorMessage();
        return;
    }

    pid = fork();
    if (pid == -1)
    {
        close(wrp[0]);
        close(wrp[1]);
        close(rwp[0]);
        close(rwp[1]);
        RunErrorMessage();
        return;
    }

    if (pid == 0)
    {
        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);
        if (dup2(wrp[0], STDIN_FILENO) == -1
                || dup2(rwp[1], STDOUT_FILENO) == -1
                || dup2(rwp[1], STDERR_FILENO) == -1)
            exit(1);

        close(rwp[0]);
        close(rwp[1]);
        close(wrp[0]);
        close(wrp[1]);

        char * argv[9];
        char day[3];
        char month[4];
        char year[5];
        char before[16];
        char after[16];

        sprintf(day, "%d", dt.DayOfMonth);
        sprintf(month, "%s", ShortMonthNames[dt.Month - 1]);
        sprintf(year, "%d", dt.Year);
        sprintf(before, "%d", b);
        sprintf(after, "%d", a);

        argv[0] = RunRMEPath;
        argv[1] = "-b";
        argv[2] = before;
        argv[3] = "-a";
        argv[4] = after;
        argv[5] = day;
        argv[6] = month;
        argv[7] = year;
        argv[8] = 0;

        execv(RunRMEPath, argv);
        exit(1);
    }

    close(rwp[1]);
    close(wrp[0]);

    UIClearText();

    char rb[1024];
    int ret;

    for (;;)
    {
        ret = read(rwp[0], rb, sizeof(rb));
        if (ret < 1)
            break;

        rb[ret] = 0;
        UIAppendText(rb);
    }

    close(rwp[0]);
    close(wrp[1]);

    int ws;

    wait(&ws);
}

static pascal OSStatus WindowEventHandler(EventHandlerCallRef nh,
    EventRef er, void * ud)
{
    HICommand hic;

    if (GetEventClass(er) == kEventClassCommand
            && GetEventKind(er) == kEventCommandProcess)
    {
        GetEventParameter(er, kEventParamDirectObject, typeHICommand, 0,
                sizeof(HICommand), 0, &hic);

        switch (hic.commandID)
        {
        case 'tDay': EngineToday(); return(noErr);
        case 'pDay': EnginePrevious(); return(noErr);
        case 'nDay': EngineNext(); return(noErr);
        case 'uPdt': EngineUpdate(); return(noErr);
        }
    }

    return(eventNotHandledErr);
}

static void UpdateBeforeAfter(ControlRef cr, char * s,
    const ControlID * cid)
{
    char buf[128];
    CFStringRef sr;

    sprintf(buf, "%d days %s", (int) GetControl32BitValue(cr), s);

    GetControlByID(Window, cid, &cr);
    sr = CFStringCreateWithBytes(0, (const UInt8 *) buf, strlen(buf),
            kCFStringEncodingMacRoman, 0);
    SetControlData(cr, kControlEntireControl,
            kControlStaticTextCFStringTag, sizeof(CFStringRef), &sr);
    CFRelease(sr);
    HIViewSetNeedsDisplay(cr, 1);
}

static pascal void BeforeHandler(ControlRef cr, SInt16 cp)
{
    UpdateBeforeAfter(cr, "before", &BeforeStaticID);
}

static pascal void AfterHandler(ControlRef cr, SInt16 cp)
{
    UpdateBeforeAfter(cr, "after", &AfterStaticID);
}

void SetupControls()
{
    HIViewRef tvr;
    HIViewRef pvr;
    HIRect rect;
    HIViewRef svr;

    HIViewFindByID(HIViewGetRoot(Window), ViewControlID, &tvr);
    pvr = HIViewGetSuperview(tvr);
    HIViewGetFrame(tvr, &rect);
    HIViewRemoveFromSuperview(tvr);

    HIScrollViewCreate(kHIScrollViewOptionsVertScroll, &svr);
    HIViewAddSubview(pvr, svr);
    HIViewSetFrame(svr, &rect);
    HIViewSetVisible(svr, 1);

    TextView = tvr;
    HIViewAddSubview(svr, TextView);
    HIViewPlaceInSuperviewAt(TextView, 0, 0);
    HIViewSetVisible(TextView, 1);

    TextObject = HITextViewGetTXNObject(TextView);

    ControlRef cr;

    GetControlByID(Window, &BeforeSliderID, &cr);
    SetControlAction(cr, NewControlActionUPP(BeforeHandler));
    UpdateBeforeAfter(cr, "before", &BeforeStaticID);

    GetControlByID(Window, &AfterSliderID, &cr);
    SetControlAction(cr, NewControlActionUPP(AfterHandler));
    UpdateBeforeAfter(cr, "after", &AfterStaticID);
}

static const EventTypeSpec WindowEvents[] =
{
    {kEventClassCommand, kEventCommandProcess}
};

int main(int argc, char * argv[])
{
    IBNibRef nr;
    char * s;

    if (realpath(argv[0], RunRMEPath) == 0)
        strcpy(RunRMEPath, argv[0]);

    s = RunRMEPath + strlen(RunRMEPath);
    while (*s != '/')
    {
        *s = 0;
        s -= 1;
    }
    strcat(RunRMEPath, "runrme");

    if (CreateNibReference(CFSTR("remindme"), &nr) != noErr)
        return(1);

    if (SetMenuBarFromNib(nr, CFSTR("AppMenu")) != noErr)
        return(1);

    if (CreateWindowFromNib(nr, CFSTR("AppWindow"), &Window) != noErr)
        return(1);

    DisposeNibReference(nr);

    InstallWindowEventHandler(Window,
            NewEventHandlerUPP(WindowEventHandler),
            sizeof(WindowEvents) / sizeof(EventTypeSpec), WindowEvents,
            Window, 0);

    SetupControls();

    EngineToday();
    EngineUpdate();

    ShowWindow(Window);
    RunApplicationEventLoop();
    return(0);
}
