/*

RemindMe

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "remindme.hpp"
#include "date.hpp"
#include "tscm.hpp"

#ifdef _WIN32_WINNT
#define strcasecmp(s1, s2) stricmp(s1, s2)
#endif // _WIN32_WINNT

static TSWorld * ScmWorld = 0;

typedef unsigned int ScmDate;

/*
ScmDate: xxddddddddmmmmyyyyyyyyyyyyyyyy11
*/

#define YearMask 0xFFFF
#define YearShift 2
#define MonthMask 0xF
#define MonthShift 18
#define DayMask 0xFF
#define DayShift 22

#define DateP(obj) ((((TSFixnum) (obj)) & 0x3) == 0x3)
#define AsDay(obj) \
    ((unsigned char) ((((ScmDate) (obj)) >> DayShift) & DayMask))
#define AsMonth(obj) \
    ((unsigned char) ((((ScmDate) (obj)) >> MonthShift) & MonthMask))
#define AsYear(obj) \
    ((unsigned int) ((((ScmDate) (obj)) >> YearShift) & YearMask))

static TSObject * MakeDate(unsigned char dom, unsigned char m, unsigned int y)
{
    return((TSObject *) ((dom << DayShift) | (m << MonthShift)
            | (y << YearShift) | 0x3));
}

int ChStringLength(Ch * s)
{
    int sl = 0;

    while (*s != 0)
    {
        s += 1;
        sl += 1;
    }

    return(sl);
}

Ch * ChStringDup(Ch * s)
{
    int sl = ChStringLength(s) + 1;
    Ch * ns = new Ch[sl];
    int sdx;

    for (sdx = 0; sdx < sl; sdx++)
        ns[sdx] = s[sdx];

    return(ns);
}

struct RMErrorType
{
    Ch * Message;
    Ch * Source;
    int LineNumber;
    int Fatal;
};

static RMErrorType Error;
void ThrowError(char * m, Ch * s, int ln, int f)
{
    Error.Message = (Ch *) m;
    Error.Source = s;
    Error.LineNumber = ln;
    Error.Fatal = f;

    throw Error;
}

#define ValueTestSpecifier 0
#define ValueSetSpecifier 1

struct RMSymbolType
{
    RMSymbolType(Ch * s, RMSymbolType * st);

    Ch * String;
    unsigned int Syntax;
// GC: need to protect Procedure
    TSObject * Procedure;
    RMSymbolType * Next;
};

RMSymbolType::RMSymbolType(Ch * s, RMSymbolType * st)
{
    String = s;
    Syntax = 0;
    Procedure = ScmWorld->Unspecified;
    Next = st;
}

#define SYMBOLTABLE 67
static RMSymbolType * SymbolTable[SYMBOLTABLE];

static unsigned int Hash(Ch * s)
{
    const char * p;
    unsigned int h = 0;
    int sl = ChStringLength(s) * sizeof(Ch);

    for (p = (const char *) s; sl > 0; p++, sl--)
        h = (h << 5) - h + *p;

    return(h);
}

void InitSymbolTable()
{
    int idx;

    for (idx = 0; idx < SYMBOLTABLE; idx++)
        SymbolTable[idx] = 0;
}

RMSymbolType * InternSymbol(Ch * s)
{
    int idx = Hash(s) % SYMBOLTABLE;
    RMSymbolType * st = SymbolTable[idx];

    while (st != 0)
    {
        if (ChStringEqualIC(s, st->String))
            return(st);

        st = st->Next;
    }

    st = new RMSymbolType(ChStringDup(s), SymbolTable[idx]);
    SymbolTable[idx] = st;

    return(st);
}

RMSymbolType * InternSymbol(char * s)
{
    Ch buf[128];
    int sdx = 0;

    while (s[sdx] != 0)
    {
        buf[sdx] = s[sdx];
        sdx += 1;
    }
    buf[sdx] = 0;

    return(InternSymbol(buf));
}

static RMSymbolType * SymbolOn;
static RMSymbolType * SymbolOr;
static RMSymbolType * SymbolBefore;
static RMSymbolType * SymbolAfter;
static RMSymbolType * SymbolPrevious;
static RMSymbolType * SymbolAnd;
static RMSymbolType * SymbolEvery;
static RMSymbolType * SymbolUntil;
static RMSymbolType * SymbolFor;
static RMSymbolType * SymbolExcept;

enum RMTokenTagType
{
    TokenSymbol,
    TokenString,
    TokenNumber,
    TokenScheme,
    TokenDot
};

struct RMTokenType
{
    RMTokenTagType Tag;
    Ch * Source;
    int LineNumber;
    union
    {
        RMSymbolType * Symbol;
        Ch * String;
        int Number;
    };
};

struct RMStatementType
{
    RMStatementType(int tc, RMTokenType * tv);
    ~RMStatementType();

    int TokenCount;
    RMTokenType * Tokens;
};

struct RMPreviousType
{
    RMPreviousType() {Count = 0; Step = 0;}

    int Count;
    int Step;
};

class RMSpecifierType
{
public:

    int Match(RMDateType dt, int npf);
    void SetPrevious(int n, int s) {Previous.Count = n; Previous.Step = s;}
    virtual RMDateType GetBaseDate(RMDateType cdt) {return(cdt);}

    RMDateType MatchDate;

protected:

    friend class RMAndSpecifierType;

    virtual int Match(RMDateType dt) = 0;

private:

    RMPreviousType Previous;
};

int RMSpecifierType::Match(RMDateType dt, int npf)
{
    MatchDate = dt;

    if (Match(dt))
        return(1);

    if (npf != 0 || Previous.Count == 0)
        return(0);

    for (int idx = 0; idx < Previous.Count; idx++)
    {
        dt.PlusDays(Previous.Step);
        if (Match(dt))
        {
            MatchDate = dt;
            return(1);
        }
    }

    return(0);
}

class RMModifierType
{
public:

    virtual ~RMModifierType() { /* Nothing. */ }
    virtual int Match(RMDateType sdt, RMDateType cdt) = 0;
};

enum RMDoneTagType
{
    DoneNone,
    DoneUntil,
    DoneFor
};

struct RMExceptType
{
    RMExceptType(RMDateType dt, RMExceptType * et) {Date = dt; Next = et;}

    RMDateType Date;
    RMExceptType * Next;
};

class RMEveryModifierType : public RMModifierType
{
public:

    RMEveryModifierType(int n, int s);
    virtual ~RMEveryModifierType();
    virtual int Match(RMDateType sdt, RMDateType cdt);
    void SetDoneUntil(RMDateType dt) {DoneTag = DoneUntil; UntilDate = dt;}
    void SetDoneFor(int n, int s)
        {DoneTag = DoneFor; ForCount = n; ForStep = s;}
    void AddExcept(RMDateType dt);

private:

    int ConfirmMatch(RMDateType sdt, RMDateType cdt);
    int CheckExcepts(RMDateType cdt);

    int Count;
    int Step;

    RMDoneTagType DoneTag;
    RMDateType UntilDate;
    int ForCount;
    int ForStep;
    RMExceptType * Excepts;
};

RMEveryModifierType::RMEveryModifierType(int n, int s)
{
    Count = n;
    Step = s;
    DoneTag = DoneNone;
    Excepts = 0;
}

RMEveryModifierType::~RMEveryModifierType()
{
    RMExceptType * et;

    while (Excepts != 0)
    {
        et = Excepts;
        Excepts = Excepts->Next;
        delete et;
    }
}

#define SyntaxValueMonth 0x0000FFFE
#define SyntaxValueYear  0x0000FFFF

int RMEveryModifierType::Match(RMDateType sdt, RMDateType cdt)
{
    if (sdt.Match(cdt))
        return(ConfirmMatch(sdt, cdt));

    if (Step == SyntaxValueYear)
    {
        if (sdt.Year == 0 || (cdt.Year > sdt.Year
                && (cdt.Year - sdt.Year) % Count == 0))
        {
            if (sdt.Year == 0)
                sdt.Year = cdt.Year;

            if (sdt.DayOfMonth == cdt.DayOfMonth && sdt.Month == cdt.Month)
                return(ConfirmMatch(sdt, cdt));
        }
    }
    else if (Step == SyntaxValueMonth)
    {
        if (sdt.Year == 0)
            sdt.Year = cdt.Year;

        if (sdt.DayOfMonth == cdt.DayOfMonth
                && sdt.GregorianMonths() < cdt.GregorianMonths()
                && (cdt.GregorianMonths() - sdt.GregorianMonths()) % Count == 0)
            return(ConfirmMatch(sdt, cdt));
    }
    else
    {
        if (sdt.Year == 0)
            sdt.Year = cdt.Year;

        unsigned int sj = sdt.Gregorian();
        unsigned int cj = cdt.Gregorian();
        if (sj < cj && (cj - sj) % (Count * Step) == 0)
            return(ConfirmMatch(sdt, cdt));
    }

    return(0);
}

void RMEveryModifierType::AddExcept(RMDateType dt)
{
    Excepts = new RMExceptType(dt, Excepts);
}

int RMEveryModifierType::ConfirmMatch(RMDateType sdt, RMDateType cdt)
{
    if (DoneTag == DoneNone)
        return(CheckExcepts(cdt));

    if (DoneTag == DoneUntil)
    {
        RMDateType dt = UntilDate;

        if (dt.Year == 0)
            dt.Year = cdt.Year;

        if (cdt.Year < dt.Year || (cdt.Year == dt.Year
            && (cdt.Month < dt.Month || (cdt.Month == dt.Month
            && cdt.DayOfMonth <= dt.DayOfMonth))))
            return(CheckExcepts(cdt));
        return(0);
    }

    // DoneTag == DoneFor

    if (ForStep == SyntaxValueYear)
    {
        if (sdt.Year + ForCount > cdt.Year)
            return(CheckExcepts(cdt));
        return(0);
    }
    else if (ForStep == SyntaxValueMonth)
    {
        if (sdt.GregorianMonths() + ForCount > cdt.GregorianMonths())
            return(CheckExcepts(cdt));
        return(0);
    }

    if (sdt.Gregorian() + ForStep * ForCount - 1 >= cdt.Gregorian())
        return(CheckExcepts(cdt));
    return(0);
}

int RMEveryModifierType::CheckExcepts(RMDateType cdt)
{
    RMExceptType * et = Excepts;

    while (et != 0)
    {
        if ((et->Date.Year == 0 || et->Date.Year == cdt.Year)
                && et->Date.Month == cdt.Month
                && et->Date.DayOfMonth == cdt.DayOfMonth)
            return(0);

        et = et->Next;
    }

    return(1);
}

class RMOffsetModifierType : public RMModifierType
{
public:

    RMOffsetModifierType(unsigned char wds, int of, int bf);
    virtual int Match(RMDateType sdt, RMDateType cdt);

private:

    RMDateType DateOffset(RMDateType dt);

    unsigned char WeekDaySet;
    int OnFlag;
    int BeforeFlag; // otherwise, after

    unsigned int YearCached;
    int CachedDateCount;
    RMDateType CachedDates[13];
};

RMOffsetModifierType::RMOffsetModifierType(unsigned char wds, int of, int bf)
{
    WeekDaySet = wds;
    OnFlag = of;
    BeforeFlag = bf;

    YearCached = 0;
    CachedDateCount = 0;
}

int RMOffsetModifierType::Match(RMDateType sdt, RMDateType cdt)
{
    RMDateType dt = sdt;

    if (sdt.DayOfMonth == 0)
        return(0);

    for (;;)
    {
        if (YearCached == cdt.Year)
        {
            for (int ddx = 0; ddx < CachedDateCount; ddx++)
                if (CachedDates[ddx].Match(cdt))
                    return(1);

            break;
        }

        CachedDateCount = 0;
        YearCached = cdt.Year;

        dt = sdt;
        if (dt.Year == 0)
            dt.Year = cdt.Year;

        if (sdt.Month == 0)
        {

            for (dt.Month = 0; dt.Month <= 12; dt.Month++)
            {
                CachedDates[CachedDateCount] = DateOffset(dt);
                CachedDateCount += 1;
            }
        }
        else
        {
            CachedDates[CachedDateCount] = DateOffset(dt);
            CachedDateCount += 1;
        }

        if (sdt.Year == 0 && (sdt.Month == 0 || sdt.Month == 12)
                && sdt.DayOfMonth > 24 && BeforeFlag == 0)
        {
            dt.Year = cdt.Year - 1;
            dt.Month = 12;
            CachedDates[CachedDateCount] = DateOffset(dt);
            CachedDateCount += 1;
        }
        else if (sdt.Year == 0 && (sdt.Month == 0 || sdt.Month == 1)
                && sdt.DayOfMonth < 7 && BeforeFlag != 0)
        {
            dt.Year = cdt.Year + 1;
            dt.Month = 1;
            CachedDates[CachedDateCount] = DateOffset(dt);
            CachedDateCount += 1;
        }
    }

    return(0);
}

RMDateType RMOffsetModifierType::DateOffset(RMDateType dt)
{
    if (BeforeFlag != 0)
    {
        if (OnFlag == 0)
            dt.Yesterday();

        for (;;)
        {
            dt.DayOfWeek = DayOfWeek(dt.DayOfMonth, dt.Month, dt.Year);
            if ((1 << dt.DayOfWeek) & WeekDaySet)
                return(dt);
            dt.Yesterday();
        }
    }

    if (OnFlag == 0)
        dt.Tomorrow();

    for (;;)
    {
        dt.DayOfWeek = DayOfWeek(dt.DayOfMonth, dt.Month, dt.Year);
        if ((1 << dt.DayOfWeek) & WeekDaySet)
            return(dt);
        dt.Tomorrow();
    }
}

class RMNthModifierType : public RMModifierType
{
public:

    RMNthModifierType(unsigned char nth, unsigned char dow);
    virtual int Match(RMDateType sdt, RMDateType cdt);

private:

    unsigned char Nth;
    unsigned char DayOfWeek;
};

RMNthModifierType::RMNthModifierType(unsigned char nth, unsigned char dow)
{
    Nth = nth;
    DayOfWeek = dow;
}

int RMNthModifierType::Match(RMDateType sdt, RMDateType cdt)
{
    if (sdt.Match(cdt)
            && NthDayOfMonth(Nth, DayOfWeek, cdt.Month, cdt.Year)
            == cdt.DayOfMonth)
        return(1);
    return(0);
}

class RMDateSpecifierType : public RMSpecifierType
{
public:

    RMDateSpecifierType(RMModifierType * mt, RMDateType dt);
    virtual int Match(RMDateType dt);
    virtual RMDateType GetBaseDate(RMDateType cdt);

private:

    RMModifierType * Modifier;
    RMDateType Date;
};

RMDateSpecifierType::RMDateSpecifierType(RMModifierType * mt,
    RMDateType dt)
{
    Modifier = mt;
    Date = dt;
}

int RMDateSpecifierType::Match(RMDateType dt)
{
    if (Modifier != 0)
        return(Modifier->Match(Date, dt));

    return(Date.Match(dt));
}

RMDateType RMDateSpecifierType::GetBaseDate(RMDateType cdt)
{
    RMDateType dt = Date;

    if (dt.DayOfMonth == 0)
        dt.DayOfMonth = cdt.DayOfMonth;
    if (dt.Month == 0)
        dt.Month = cdt.Month;
    if (dt.Year == 0)
        dt.Year = cdt.Year;

    return(dt);
}

class RMAndSpecifierType : public RMSpecifierType
{
public:

    RMAndSpecifierType(int sptc, RMSpecifierType ** sptv);
    virtual int Match(RMDateType dt);

private:

    int Count;
    RMSpecifierType ** Specifiers;
};

RMAndSpecifierType::RMAndSpecifierType(int sptc, RMSpecifierType ** sptv)
{
    int idx;

    Count = sptc;
    Specifiers = new RMSpecifierType * [Count];
    for (idx = 0; idx < Count; idx++)
        Specifiers[idx] = sptv[idx];
}

int RMAndSpecifierType::Match(RMDateType dt)
{
    int idx;

    for (idx = 0; idx < Count; idx++)
        if (Specifiers[idx]->Match(dt) != 0)
            return(1);

    return(0);
}

class RMActionType
{
public:

    virtual ~RMActionType() { /* Nothing. */ }
    virtual void Trigger(RMDateType cdt, RMDateType bdt, RMDateType mdt) = 0;
};

enum RMActionTagType
{
    ActionRemind
};

class RMActionRemind : public RMActionType
{
public:

    RMActionRemind(Ch * s) {String = ChStringDup(s);}
    virtual ~RMActionRemind() {delete String;}
    virtual void Trigger(RMDateType cdt, RMDateType bdt, RMDateType mdt);

private:

    void EvaluateField(Ch * fn, RMDateType cdt, RMDateType bdt,
            RMDateType mdt);

    Ch * String;
};

struct RMReminderType
{
    RMReminderType(RMSpecifierType * st, int ac, RMActionType ** av);
    ~RMReminderType();

    RMSpecifierType * Specifier;
    int ActionCount;
    RMActionType ** Actions;
};

class InputStream
{
public:

    virtual ~InputStream() { /* Nothing. */ }
    virtual Ch * GetFilename() = 0;
    virtual Ch * GetPath() = 0;
    virtual Ch GetCh(int * eof) = 0;
};

char * ShortMonth[13] = {0, "jan", "feb", "mar", "apr", "may", "jun", "jul",
    "aug", "sep", "oct", "nov", "dec"};
char * ShortDayOfWeek[8] = {0, "sun", "mon", "tue", "wed", "thu", "fri",
    "sat"};

#define SyntaxWeekDay         0x80000000
#define SyntaxMonth           0x40000000
#define SyntaxAction          0x20000000
#define SyntaxNth             0x10000000
#define SyntaxPreviousStep    0x08000000
#define SyntaxEveryStep       0x04000000
#define SyntaxSpecifier       0x02000000
#define SyntaxField           0x01000000

#define SyntaxMask    0x0000FFFF

struct InitSymbol
{
    char * String;
    unsigned int Syntax;
};

const static InitSymbol InitSymbols[] =
{
    {"sun", SyntaxWeekDay | 1},
    {"sunday", SyntaxWeekDay | 1},
    {"mon", SyntaxWeekDay | 2},
    {"monday", SyntaxWeekDay | 2},
    {"tue", SyntaxWeekDay | 3},
    {"tuesday", SyntaxWeekDay | 3},
    {"wed", SyntaxWeekDay | 4},
    {"wednesday", SyntaxWeekDay | 4},
    {"thu", SyntaxWeekDay | 5},
    {"thursday", SyntaxWeekDay | 5},
    {"fri", SyntaxWeekDay | 6},
    {"friday", SyntaxWeekDay | 6},
    {"sat", SyntaxWeekDay | 7},
    {"saturday", SyntaxWeekDay | 7},

    {"jan", SyntaxMonth | 1},
    {"january", SyntaxMonth | 1},
    {"feb", SyntaxMonth | 2},
    {"february", SyntaxMonth | 2},
    {"mar", SyntaxMonth | 3},
    {"march", SyntaxMonth | 3},
    {"apr", SyntaxMonth | 4},
    {"april", SyntaxMonth | 4},
    {"may", SyntaxMonth | 5},
    {"jun", SyntaxMonth | 6},
    {"june", SyntaxMonth | 6},
    {"jul", SyntaxMonth | 7},
    {"july", SyntaxMonth | 7},
    {"aug", SyntaxMonth | 8},
    {"august", SyntaxMonth | 8},
    {"sep", SyntaxMonth | 9},
    {"september", SyntaxMonth | 9},
    {"oct", SyntaxMonth | 10},
    {"october", SyntaxMonth | 10},
    {"nov", SyntaxMonth | 11},
    {"november", SyntaxMonth | 11},
    {"dec", SyntaxMonth | 12},
    {"december", SyntaxMonth | 12},

    {"remind", SyntaxAction | ActionRemind},

    {"first", SyntaxNth | NthFirst},
    {"second", SyntaxNth | NthSecond},
    {"third", SyntaxNth | NthThird},
    {"fourth", SyntaxNth | NthFourth},
    {"fifth", SyntaxNth | NthFifth},
    {"last", SyntaxNth | NthLast},

    {"day", SyntaxPreviousStep | SyntaxEveryStep | 1},
    {"days", SyntaxPreviousStep | SyntaxEveryStep | 1},
    {"week", SyntaxPreviousStep | SyntaxEveryStep | 7},
    {"weeks", SyntaxPreviousStep | SyntaxEveryStep | 7},
    {"month", SyntaxEveryStep | SyntaxValueMonth},
    {"months", SyntaxEveryStep | SyntaxValueMonth},
    {"year", SyntaxEveryStep | SyntaxValueYear},
    {"years", SyntaxEveryStep | SyntaxValueYear}
};

class ScmOutputStream : public TSOutputStream
{
public:

    ScmOutputStream(FILE * f) {File = f;}
    virtual void PutCh(TSCh ch) {fputc(ch, File);}
    virtual void PutString(TSCh * s, int sl);
    virtual void PutString(char * s) {fputs(s, File);}

private:

    FILE * File;
};

void ScmOutputStream::PutString(TSCh * s, int sl)
{
    int sdx;

    for (sdx = 0; sdx < sl; sdx++)
        PutCh(s[sdx]);
}

void MustBeDate(TSWorld * wld, TSObject * obj)
{
    if (DateP(obj) == 0)
        wld->ThrowError("Wrong type", obj);
}

void RMActionRemind::Trigger(RMDateType cdt, RMDateType bdt, RMDateType mdt)
{
/*    printf("%d.%d.%d %d.%d.%d %d.%d.%d: ",
            cdt.DayOfMonth, cdt.Month, cdt.Year,
            bdt.DayOfMonth, bdt.Month, bdt.Year,
            mdt.DayOfMonth, mdt.Month, mdt.Year);
*/

    Ch * s = String;
    Ch fn[128];
    int fnl;

    while (*s != 0)
    {
        if (*s == '%')
        {
            s += 1;
            if (*s == '(')
            {
                s += 1;
                fnl = 0;
                while (*s != ')' && *s != 0)
                {
                    fn[fnl] = *s;
                    s += 1;
                    fnl += 1;
                    if (fnl + 1 == sizeof(fn) / sizeof(Ch))
                        break;
                }

                if (*s == ')')
                    s += 1;

                fn[fnl] = 0;
                EvaluateField(fn, cdt, bdt, mdt);
            }
            else
                putc('%', stdout);
        }
        else
        {
            putc(*s, stdout);
            s += 1;
        }
    }

    putc('\n', stdout);
}

#define AsDate(dt) MakeDate((dt).DayOfMonth, (dt).Month, (dt).Year)

void RMActionRemind::EvaluateField(Ch * fn, RMDateType cdt, RMDateType bdt,
    RMDateType mdt)
{
    RMSymbolType * st = InternSymbol(fn);

    if (st->Syntax & SyntaxField)
    {
        TSObject * av[3];
        av[0] = AsDate(cdt);
        av[1] = AsDate(bdt);
        av[2] = AsDate(mdt);
        ScmWorld->Display(ScmWorld->DefaultOutput, Evaluate(ScmWorld,
                st->Procedure->Call(ScmWorld, 3, av)));
    }
}

RMDateType AsDateType(TSObject * obj)
{
    RMDateType dt;

    dt.DayOfMonth = AsDay(obj);
    dt.Month = AsMonth(obj);
    dt.Year = AsYear(obj);

    return(dt);
}

class RMTestSpecifierType : public RMSpecifierType
{
public:

    RMTestSpecifierType(TSObject * proc) {Procedure = proc;}
    virtual int Match(RMDateType dt);

private:

// GC: need to protect Procedure

    TSObject * Procedure;
};

int RMTestSpecifierType::Match(RMDateType dt)
{
    TSObject * av[1];
    av[0] = AsDate(dt);
    if (Evaluate(ScmWorld, Procedure->Call(ScmWorld, 1, av))
            == ScmWorld->False)
        return(0);
    return(1);
}

class RMSetSpecifierType : public RMSpecifierType
{
public:

    RMSetSpecifierType(TSObject * proc)
        {Procedure = proc; Year = 0; Set = ScmWorld->EmptyList;}
    virtual int Match(RMDateType dt);

private:

// GC: need to protect Procedure and Set

    TSObject * Procedure;
    unsigned int Year;
    TSObject * Set;
};

int RMSetSpecifierType::Match(RMDateType dt)
{
    TSObject * obj;
    TSObject * d = AsDate(dt);

    if (dt.Year != Year)
    {
        TSObject * av[1];
        av[0] = ScmWorld->MakeNumber(dt.Year);
        Set = Evaluate(ScmWorld, Procedure->Call(ScmWorld, 1, av));
        Year = dt.Year;
    }

    for (obj = Set; obj != ScmWorld->EmptyList; obj = Rest(obj))
    {
        ScmWorld->MustBe(PairType, obj);

        if (First(obj) == d)
            return(1);
    }

    return(0);
}

TSObject * ScmDayNames[7];

class RMApplication : public TSApplication
{
public:

    virtual void FatalError();
    virtual int Write(TSOutputStream * os, TSObject * obj);
    virtual int Display(TSOutputStream * os, TSObject * obj) {return(0);}
};

void RMApplication::FatalError()
{
    exit(1);
}

int RMApplication::Write(TSOutputStream * os, TSObject * obj)
{
    if (DateP(obj))
    {
        char buf[128];

        sprintf(buf, "#<date: %d %s %d>", AsDay(obj), ShortMonth[AsMonth(obj)],
                AsYear(obj));
        os->PutString(buf);
        return(1);
    }

    return(0);
}

void InitRemindMe()
{
    unsigned int idx;

    InitSymbolTable();

    ScmWorld = new TSWorld(new ScmOutputStream(stdout), new RMApplication);
    AddAllBuiltins(ScmWorld);

    ScmDayNames[0] = ScmWorld->InternSymbol("sunday");
    ScmDayNames[1] = ScmWorld->InternSymbol("monday");
    ScmDayNames[2] = ScmWorld->InternSymbol("tuesday");
    ScmDayNames[3] = ScmWorld->InternSymbol("wednesday");
    ScmDayNames[4] = ScmWorld->InternSymbol("thursday");
    ScmDayNames[5] = ScmWorld->InternSymbol("friday");
    ScmDayNames[6] = ScmWorld->InternSymbol("saturday");

    for (idx = 0; idx < sizeof(InitSymbols) / sizeof(InitSymbol); idx++)
        InternSymbol(InitSymbols[idx].String)->Syntax =
                InitSymbols[idx].Syntax;

    SymbolOn = InternSymbol("on");
    SymbolOr = InternSymbol("or");
    SymbolBefore = InternSymbol("before");
    SymbolAfter = InternSymbol("after");
    SymbolPrevious = InternSymbol("previous");
    SymbolAnd = InternSymbol("and");
    SymbolEvery = InternSymbol("every");
    SymbolUntil = InternSymbol("until");
    SymbolFor = InternSymbol("for");
    SymbolExcept = InternSymbol("except");
}

RMStatementType::RMStatementType(int tc, RMTokenType * tv)
{
    int tdx;

    TokenCount = tc;
    Tokens = new RMTokenType[TokenCount];
    for (tdx = 0; tdx < TokenCount; tdx++)
        Tokens[tdx] = tv[tdx];
}

RMStatementType::~RMStatementType()
{
    int tdx;

    if (Tokens != 0)
    {
        for (tdx = 0; tdx < TokenCount; tdx++)
            if (Tokens[tdx].Tag == TokenString)
                delete Tokens[tdx].String;
        delete Tokens;
    }
}

RMReminderType::RMReminderType(RMSpecifierType * st, int ac,
    RMActionType ** av)
{
    int adx;

    Specifier = st;
    ActionCount = ac;
    Actions = new RMActionType * [ActionCount];
    for (adx = 0; adx < ActionCount; adx++)
        Actions[adx] = av[adx];
}

RMReminderType::~RMReminderType()
{
    int adx;

    delete Specifier;
    if (Actions != 0)
    {
        for (adx = 0; adx < ActionCount; adx++)
            delete Actions[adx];
        delete Actions;
    }
}

class ParseInputStream
{
public:

    ParseInputStream(InputStream * is, ParseInputStream * pis);
    ~ParseInputStream() {delete Input;}
    Ch * GetFilename() {return(Input->GetFilename());}
    Ch * GetPath() {return(Input->GetPath());}
    int GetLineNumber() {return(LineNumber);}
    Ch GetCh(int * eof);
    void UnGetCh(Ch ch);
    ParseInputStream * GetPrevious() {return(Previous);}

private:

    InputStream * Input;
    Ch NextCh;
    Ch LastCh;
    char NextChValid;
    int LineNumber;
    ParseInputStream * Previous;
};

class ScmInputStream : public TSInputStream
{
public:

    ScmInputStream(ParseInputStream * pis) {Input = pis; Filename = 0;}
    virtual ~ScmInputStream();
    virtual TSCh GetCh(int * eof) {return((TSCh) Input->GetCh(eof));}
    virtual void UnGetCh(TSCh ch) {Input->UnGetCh((Ch) ch);}
    virtual TSCh * GetFilename();
    virtual int GetLineNumber() {return(Input->GetLineNumber());}

private:

    ParseInputStream * Input;
    TSCh * Filename;
};

ScmInputStream::~ScmInputStream()
{
    if (Filename != 0)
        delete Filename;
}

TSCh * ScmInputStream::GetFilename()
{
    if (Filename == 0)
    {
        Ch * s = Input->GetFilename();
        int sl = ChStringLength(s);
        int sdx;

        Filename = new TSCh[sl + 1];
        for (sdx = 0; sdx <= sl; sdx++)
            Filename[sdx] = s[sdx];
    }

    return(Filename);
}

ParseInputStream::ParseInputStream(InputStream * is, ParseInputStream * pis)
{
    Input = is;
    NextCh = 0;
    LastCh = 0;
    NextChValid = 0;
    LineNumber = 1;
    Previous = pis;
}

Ch ParseInputStream::GetCh(int * eof)
{
    if (NextChValid)
    {
        *eof = 0;
        NextChValid = 0;
        return(NextCh);
    }

    LastCh = NextCh;
    NextCh = Input->GetCh(eof);
    if (*eof == 0)
    {
        if (NextCh == 13)
            LineNumber += 1;
        else if (NextCh == 10 && LastCh != 13)
            LineNumber += 1;
    }

    return(NextCh);
}

void ParseInputStream::UnGetCh(Ch ch)
{
    NextChValid = 1;
    NextCh = ch;
}

class RMReminderStream
{
public:

    RMReminderStream(InputStream * is);
    ~RMReminderStream();
    RMReminderType * ParseReminder();
    void Include(InputStream * is);
    ParseInputStream * GetParseInputStream() {return(ParseStream);}

private:

    int ParseToken(RMTokenType * tt);
    void ParseExpression(int ef);
    RMStatementType * ParseStatement();
    int ParseDate(RMStatementType * st, int tdx, RMDateType * dt);
    int ParseDateSpecifier(RMStatementType * st, RMSpecifierType ** spt,
            int tdx);
    RMReminderType * ParseReminder(RMStatementType * st);

    ParseInputStream * ParseStream;
};

RMReminderStream::RMReminderStream(InputStream * is)
{
    ParseStream = new ParseInputStream(is, 0);
}

RMReminderStream::~RMReminderStream()
{
    ParseInputStream * pis;

    while (ParseStream != 0)
    {
        pis = ParseStream;
        ParseStream = ParseStream->GetPrevious();
        delete pis;
    }
}

RMReminderType * RMReminderStream::ParseReminder()
{
    RMStatementType * st = 0;

    st = ParseStatement();
    if (st == 0)
        return(0);
    return(ParseReminder(st));
}

void RMReminderStream::Include(InputStream * is)
{
    ParseStream = new ParseInputStream(is, ParseStream);
}

int RMReminderStream::ParseToken(RMTokenType * tt)
{
    Ch ch;
    Ch s[512];
    int sdx;
    int n;
    int eof;

    for (;;)
    {
        ch = ParseStream->GetCh(&eof);
        if (eof)
            return(0);
        if (ch == ';')
        {
            for (;;)
            {
                ch = ParseStream->GetCh(&eof);
                if (eof)
                    return(0);
                if (ch == 10 || ch == 13)
                    break;
            }
        }
        else if (ChLetter(ch) || ChNumber(ch) || ch == '[' || ch == '.'
                || ch == '(' || ch == '-')
            break;
    }

    tt->Source = ParseStream->GetFilename();
    tt->LineNumber = ParseStream->GetLineNumber();

    if (ChLetter(ch) || ch == '-')
    {
        sdx = 0;
        for (;;)
        {
            s[sdx] = ch;
            sdx += 1;
            if (sdx == sizeof(s) / sizeof(Ch))
                return(0);
            ch = ParseStream->GetCh(&eof);
            if (eof)
                ThrowError("unexpected end of file", tt->Source,
                        tt->LineNumber, 1);
            if (ChLetter(ch) == 0 && ChNumber(ch) == 0 && ch != '-')
                break;
        }

        ParseStream->UnGetCh(ch);
        s[sdx] = 0;
        tt->Tag = TokenSymbol;
        tt->Symbol = InternSymbol(s);
    }
    else if (ChNumber(ch))
    {
        n = ch - '0';

        for (;;)
        {
            ch = ParseStream->GetCh(&eof);
            if (eof)
                ThrowError("unexpected end of file", tt->Source,
                        tt->LineNumber, 1);
            if (ChNumber(ch) == 0)
                break;
            n = n * 10 + ch - '0';
        }

        ParseStream->UnGetCh(ch);
        tt->Tag = TokenNumber;
        tt->Number = n;
    }
    else if (ch == '[')
    {
        sdx = 0;
        for (;;)
        {
            ch = ParseStream->GetCh(&eof);
            if (ch == '\\')
                ch = ParseStream->GetCh(&eof);
            if (eof)
                ThrowError("unexpected end of file", tt->Source,
                        tt->LineNumber, 1);
            if (ch == ']')
                break;
            s[sdx] = ch;
            sdx += 1;
            if (sdx == sizeof(s) / sizeof(Ch))
                return(0);
        }

        s[sdx] = 0;
        tt->Tag = TokenString;
        tt->String = ChStringDup(s);
    }
    else if (ch == '(')
        tt->Tag = TokenScheme;
    else // ch == '.'
        tt->Tag = TokenDot;

    return(1);
}

void RMReminderStream::ParseExpression(int ef)
{
    TSObject * obj;
    ScmInputStream sis(ParseStream);
    TSLineNumberMap lnm(sis.GetFilename());

    obj = ScmWorld->Read(&sis, &lnm);
    if (obj != ScmWorld->EndOfFile && ef != 0)
    {
        obj = ScmWorld->Evaluate(obj, &lnm);
        if (obj != ScmWorld->Unspecified)
        {
            ScmWorld->Write(ScmWorld->DefaultOutput, obj);
            ScmWorld->DefaultOutput->PutString("\n");
        }
    }
}

RMStatementType * RMReminderStream::ParseStatement()
{
    RMTokenType tv[256];
    int tc = 0;

    for (;;)
    {
        while (ParseToken(tv + tc) == 0)
        {
            if (tc != 0)
                ThrowError("unexpected end of file", tv[tc - 1].Source,
                        tv[tc - 1].LineNumber, 1);

            ParseInputStream * pis = ParseStream;
            ParseStream = ParseStream->GetPrevious();
            delete pis;
            if (ParseStream == 0)
                return(0);
        }

        if (tv[tc].Tag == TokenDot)
            break;

        if (tv[tc].Tag == TokenScheme)
        {
            if (tc > 0)
                ThrowError("unexpected scheme expression", tv[tc - 1].Source,
                        tv[tc - 1].LineNumber, 0);

            ParseStream->UnGetCh('(');
            ParseExpression(tc == 0);
        }
        else
            tc += 1;

        if (tc == sizeof(tv) / sizeof(RMTokenType))
            tc -= 1;
    }

    return(new RMStatementType(tc, tv));
}

int RMReminderStream::ParseDate(RMStatementType * st, int tdx,
    RMDateType * dt)
{
/*
    <date> := [<day> | <dayname> | <month> | <year>] ...
    <day> := 1 | 2 | ... | 31
    <dayname> := mon | tue | ... | sun
            | monday | tuesday | ... | sunday
    <month> := jan | feb | ... | dec
            | january | february | ... | december
    <year> := 1800 | ... | 3000
*/

    for (; tdx < st->TokenCount; tdx++)
    {
        if (st->Tokens[tdx].Tag == TokenNumber)
        {
            int n = st->Tokens[tdx].Number;
            if (n >= 1 && n <= 31)
            {
                if (dt->DayOfMonth != 0)
                    ThrowError("day of month specified twice",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                dt->DayOfMonth = n;
            }
            else if (n >= 1800 && n <= 3000)
            {
                if (dt->Year != 0)
                    ThrowError("year specified twice",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                dt->Year = n;
            }
            else
                ThrowError("unexpected number",
                        st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);
        }
        else if (st->Tokens[tdx].Tag == TokenSymbol)
        {
            RMSymbolType * s = st->Tokens[tdx].Symbol;
            if (s->Syntax & SyntaxWeekDay)
            {
                if (dt->DayOfWeek != 0)
                    ThrowError("day of week specified twice",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                dt->DayOfWeek = s->Syntax & SyntaxMask;
            }
            else if (s->Syntax & SyntaxMonth)
            {
                if (dt->Month != 0)
                    ThrowError("month specified twice",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                dt->Month = s->Syntax & SyntaxMask;
            }
            else
                break;
        }
        else
            break;
    }

    if (dt->DayOfWeek != 0 && dt->DayOfMonth != 0)
    {
        if (dt->Month == 0 || dt->Year == 0 || DayOfWeek(dt->DayOfMonth,
                dt->Month, dt->Year) != dt->DayOfWeek)
            ThrowError("day of week and day of month conflict",
                    st->Tokens[tdx - 1].Source,
                    st->Tokens[tdx - 1].LineNumber, 0);
    }

    return(tdx);
}

int RMReminderStream::ParseDateSpecifier(RMStatementType * st,
    RMSpecifierType ** spt, int tdx)
{
    RMDateType dt;
    RMPreviousType pt;
    RMModifierType * mt = 0;
    int idx;

/*
    <date-specifier> := <date>
        | <nth-weekday> <date>
        | <offset> <date>
        | <date> <every>
        | <specifier-test>
        | <specifier-set>
*/

/*
<specifier-test>
<specifier-set>
*/

    if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenSymbol
            && st->Tokens[tdx].Symbol->Syntax & SyntaxSpecifier)
    {
        if ((st->Tokens[tdx].Symbol->Syntax & SyntaxMask)
                == ValueTestSpecifier)
            *spt = new RMTestSpecifierType(
                    st->Tokens[tdx].Symbol->Procedure);
        else //  == ValueSetSpecifier
            *spt = new RMSetSpecifierType(
                    st->Tokens[tdx].Symbol->Procedure);

        tdx += 1;
        return(tdx);
    }

/*
    <nth-weekday> := <nth> <weekday>
    <nth> := first | second | third | fourth | fifth | last
*/

    if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenSymbol
            && st->Tokens[tdx].Symbol->Syntax & SyntaxNth)
    {
        unsigned char nth = st->Tokens[tdx].Symbol->Syntax & SyntaxMask;
        tdx += 1;
        if (tdx == st->TokenCount || st->Tokens[tdx].Tag != TokenSymbol
                || (st->Tokens[tdx].Symbol->Syntax & SyntaxWeekDay) == 0)
            ThrowError("expected a week day",
                st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);
        mt = new RMNthModifierType(nth,
                st->Tokens[tdx].Symbol->Syntax & SyntaxMask);
        tdx += 1;
    }

/*
    <offset> := <weekday> <weekday> ... [on or] before
            | <weekday> <weekday> ... [on or] after
*/

    if (mt == 0)
    {
        for (idx = tdx; idx < st->TokenCount; idx++)
            if (st->Tokens[idx].Tag != TokenSymbol
                    || (st->Tokens[idx].Symbol->Syntax & SyntaxWeekDay) == 0)
                break;

        if (idx < st->TokenCount && st->Tokens[idx].Tag == TokenSymbol
                && (st->Tokens[idx].Symbol == SymbolOn
                || st->Tokens[idx].Symbol == SymbolBefore
                || st->Tokens[idx].Symbol == SymbolAfter))
        {
            unsigned char wds = 0;
            int of = 0;
            int bf = 0;

            if (idx == tdx)
                ThrowError("weekday(s) missing before on, before, or after",
                        st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);

            while (st->Tokens[tdx].Tag == TokenSymbol
                    && st->Tokens[tdx].Symbol->Syntax & SyntaxWeekDay)
            {
                wds |= (1 << (st->Tokens[tdx].Symbol->Syntax & SyntaxMask));
                tdx += 1;
            }

            if (st->Tokens[tdx].Symbol == SymbolOn)
            {
                of = 1;
                tdx += 1;
                if (tdx == st->TokenCount || st->Tokens[tdx].Tag != TokenSymbol
                        || st->Tokens[tdx].Symbol != SymbolOr)
                    ThrowError("or missing after on",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                tdx += 1;
            }

            if (tdx == st->TokenCount || st->Tokens[tdx].Tag != TokenSymbol)
                ThrowError("missing before or after",
                        st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);
            if (st->Tokens[tdx].Symbol == SymbolBefore)
                bf = 1;
            else if (st->Tokens[tdx].Symbol != SymbolAfter)
                ThrowError("missing before or after",
                        st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);
            tdx += 1;

            mt = new RMOffsetModifierType(wds, of, bf);
        }
    }

/*
<date> := ...
*/

    tdx = ParseDate(st, tdx, &dt);

/*
<every> := every [<n>] <step> [<done>] [<except> ...]
<step> := day[s] | week[s] | month[s] | year[s]
<done> := until <date>
        | for <n> <step>
<except> := except <date>
*/

    if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenSymbol
            && st->Tokens[tdx].Symbol == SymbolEvery)
    {
        int n = 1;
        int s;

        if (mt != 0)
            ThrowError("can not use every with an nth or an offset",
                    st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);

        if (dt.DayOfMonth == 0 || dt.Month == 0)
            ThrowError("must specify day and month for every",
                    st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);

        tdx += 1;
        if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenNumber)
        {
            n = st->Tokens[tdx].Number;
            tdx += 1;
        }

        if (tdx >= st->TokenCount || st->Tokens[tdx].Tag != TokenSymbol
                || ((st->Tokens[tdx].Symbol->Syntax & SyntaxEveryStep) == 0))
            ThrowError("missing step for every",
                    st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);
        s = st->Tokens[tdx].Symbol->Syntax & SyntaxMask;
        tdx += 1;

        RMEveryModifierType * emt = new RMEveryModifierType(n, s);
        mt = emt;

        if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenSymbol)
        {
            if (st->Tokens[tdx].Symbol == SymbolUntil)
            {
                RMDateType dt;

                tdx += 1;
                tdx = ParseDate(st, tdx, &dt);
                if (dt.Month == 0 || dt.DayOfMonth == 0)
                    ThrowError("must specify day and month for until",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                emt->SetDoneUntil(dt);
            }
            else if (st->Tokens[tdx].Symbol == SymbolFor)
            {
                int n, s;

                tdx += 1;
                if (tdx >= st->TokenCount
                        || st->Tokens[tdx].Tag != TokenNumber)
                    ThrowError("missing count in for",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                n = st->Tokens[tdx].Number;
                tdx += 1;

                if (tdx >= st->TokenCount || st->Tokens[tdx].Tag != TokenSymbol
                        || ((st->Tokens[tdx].Symbol->Syntax & SyntaxEveryStep)
                        == 0))
                    ThrowError("missing step in for",
                            st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber,
                            0);
                s = st->Tokens[tdx].Symbol->Syntax & SyntaxMask;
                tdx += 1;
                emt->SetDoneFor(n, s);
            }
        }

        while (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenSymbol
                && st->Tokens[tdx].Symbol == SymbolExcept)
        {
            RMDateType dt;

            tdx += 1;
            tdx = ParseDate(st, tdx, &dt);
            if (dt.Month == 0 || dt.DayOfMonth == 0)
                ThrowError("must specify day and month for except",
                        st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);
            emt->AddExcept(dt);
        }
    }

    *spt = new RMDateSpecifierType(mt, dt);
    return(tdx);
}

RMReminderType * RMReminderStream::ParseReminder(RMStatementType * st)
{
    RMSpecifierType * sptv[128];
    RMSpecifierType * spt;
    int sptc = 0;
    int tdx = 0;
    int ac;
    RMActionType * av[32];

/*
    <specifier> := <date-specifier> [and <date-specifier] ...  [<previous>]
*/

    for (;;)
    {
        tdx = ParseDateSpecifier(st, &spt, tdx);
        sptv[sptc] = spt;
        sptc += 1;

        if (tdx == st->TokenCount || st->Tokens[tdx].Tag != TokenSymbol
                || st->Tokens[tdx].Symbol != SymbolAnd)
            break;
        tdx += 1;
    }

    if (sptc == 1)
        spt = sptv[0];
    else
        spt = new RMAndSpecifierType(sptc, sptv);

/*
<previous> := previous [<n>] [day | days | week | weeks]
<n> := 1 | 2 | ...
*/

    if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenSymbol
            && st->Tokens[tdx].Symbol == SymbolPrevious)
    {
        int s = 1;
        int n = 1;

        tdx += 1;
        if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenNumber)
        {
            n = st->Tokens[tdx].Number;
            tdx += 1;
        }

        if (tdx < st->TokenCount && st->Tokens[tdx].Tag == TokenSymbol
                && (st->Tokens[tdx].Symbol->Syntax & SyntaxPreviousStep))
        {
            s = st->Tokens[tdx].Symbol->Syntax & SyntaxMask;
            tdx += 1;
        }

        spt->SetPrevious(n, s);
    }

/*
<action> := remind [<string>]
*/

    for (ac = 0; tdx < st->TokenCount; ac++)
    {
        if (ac == sizeof(av) / sizeof(RMActionType *))
            ThrowError("too many actions",
                    st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);

        if (st->Tokens[tdx].Tag != TokenSymbol
                || (st->Tokens[tdx].Symbol->Syntax & SyntaxAction) == 0)
            ThrowError("not an action",
                    st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);

        switch (st->Tokens[tdx].Symbol->Syntax & SyntaxMask)
        {
        case ActionRemind:
            tdx += 1;
            if (tdx == st->TokenCount || st->Tokens[tdx].Tag != TokenString)
                ThrowError("remind must be followed by a string",
                        st->Tokens[tdx].Source, st->Tokens[tdx].LineNumber, 0);
            av[ac] = new RMActionRemind(st->Tokens[tdx].String);
            tdx += 1;
            break;
        }
    }

    return(new RMReminderType(spt, ac, av));
}

RMReminderStream * ReminderStream = 0;

DefineBuiltin("date?", DateP, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    if (DateP(argv[0]))
        return(wld->True);

    return(wld->False);
}

DefineBuiltin("date", Date, "nnn")(TSWorld * wld, int argc, TSObject * argv[])
{
    TSFixnum d = AsFixnum(argv[0]);
    TSFixnum m = AsFixnum(argv[1]);
    TSFixnum y = AsFixnum(argv[2]);

    if (m < 1 || m > 12)
        wld->ThrowError("Month too big or too small", argv[1]);
    if (y < 1800 || y > 3000)
        wld->ThrowError("Year too big or too small", argv[2]);
    if (d < 1 || d > DaysInMonth(m, y))
        wld->ThrowError("Day too big or too small", argv[0]);

    return(MakeDate(d, m, y));
}

DefineBuiltin("day", Day, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    MustBeDate(wld, argv[0]);
    return(wld->MakeNumber(AsDay(argv[0])));
}

DefineBuiltin("month", Month, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    MustBeDate(wld, argv[0]);
    return(wld->MakeNumber(AsMonth(argv[0])));
}

DefineBuiltin("year", Year, ".")(TSWorld * wld, int argc, TSObject * argv[])
{
    MustBeDate(wld, argv[0]);
    return(wld->MakeNumber(AsYear(argv[0])));
}

DefineBuiltin("day-name", DayName, ".")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    MustBeDate(wld, argv[0]);
    return(ScmDayNames[DayOfWeek(AsDay(argv[0]), AsMonth(argv[0]),
            AsYear(argv[0])) - 1]);
}

DefineBuiltin("days-between", DaysBetween, "..")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    MustBeDate(wld, argv[0]);
    MustBeDate(wld, argv[1]);
    RMDateType dt1 = AsDateType(argv[0]);
    RMDateType dt2 = AsDateType(argv[1]);

    return(wld->MakeNumber(dt1.Gregorian() - dt2.Gregorian()));
}

DefineBuiltin("date+", DatePlus, ".n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    MustBeDate(wld, argv[0]);
    RMDateType dt = AsDateType(argv[0]);
    TSFixnum n = AsFixnum(argv[1]);
    if (n == 1)
        dt.Tomorrow();
    else if (n == -1)
        dt.Yesterday();
    else if (n > 0)
        dt.PlusDays(n);
    else
        dt.MinusDays(- n);

    return(AsDate(dt));
}

DefineBuiltin("date-", DateMinus, ".n")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    MustBeDate(wld, argv[0]);
    RMDateType dt = AsDateType(argv[0]);
    TSFixnum n = AsFixnum(argv[1]);
    if (n == -1)
        dt.Tomorrow();
    else if (n == 1)
        dt.Yesterday();
    else if (n < 0)
        dt.PlusDays(- n);
    else
        dt.MinusDays(n);

    return(AsDate(dt));
}

void ScmSpecifier(TSWorld * wld, TSObject * proc, TSObject * name, int vtt)
{
    if (ProcedureP(proc) == 0 && BuiltinP(proc) == 0)
        wld->ThrowError("Not a callable", proc);

    Ch sn[128];
    int sdx;

    for (sdx = 0; (unsigned int) sdx < sizeof(sn) / sizeof(Ch) - 1
            && sdx < AsString(name)->Length; sdx++)
        sn[sdx] = (Ch) AsString(name)->String[sdx];
    sn[sdx] = 0;

    RMSymbolType * st = InternSymbol(sn);
    st->Procedure = proc;
    st->Syntax = SyntaxSpecifier | vtt;
}

DefineBuiltin("test-specifier", TestSpecifier, ".s")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    ScmSpecifier(wld, argv[0], argv[1], ValueTestSpecifier);
    return(wld->Unspecified);
}

DefineBuiltin("set-specifier", SetSpecifier, ".s")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    ScmSpecifier(wld, argv[0], argv[1], ValueSetSpecifier);
    return(wld->Unspecified);
}

DefineBuiltin("remind-field", RemindField, ".s")(TSWorld * wld,
    int argc, TSObject * argv[])
{
    if (ProcedureP(argv[0]) == 0 && BuiltinP(argv[0]) == 0)
        wld->ThrowError("Not a callable", argv[0]);

    Ch sn[128];
    int sdx;

    for (sdx = 0; (unsigned int) sdx < sizeof(sn) / sizeof(Ch) - 1
            && sdx < AsString(argv[1])->Length; sdx++)
        sn[sdx] = (Ch) AsString(argv[1])->String[sdx];
    sn[sdx] = 0;

    RMSymbolType * st = InternSymbol(sn);
    st->Procedure = argv[0];
    st->Syntax = SyntaxField;

    return(wld->Unspecified);
}

class StdioInputStream : public InputStream
{
public:

    StdioInputStream(Ch * s, FILE * f);
    virtual ~StdioInputStream();
    virtual Ch * GetFilename() {return(Source);}
    virtual Ch * GetPath() {return(Path);}
    virtual Ch GetCh(int * eof);

private:

    Ch * Source;
    Ch * Path;
    FILE * File;
};

StdioInputStream::StdioInputStream(Ch * s, FILE * f)
{
    Source = s;
    Path = GetPathFromFilename(s);
    File = f;
}

StdioInputStream::~StdioInputStream()
{
    delete Source;
    delete Path;
    fclose(File);
}

Ch StdioInputStream::GetCh(int * eof)
{
    int ch;

    ch = fgetc(File);

    if (ch == EOF)
    {
        *eof = 1;
        return(0);
    }

    *eof = 0;
    return((Ch) ch);
}

DefineBuiltin("include", Include, "s")(TSWorld * wld, int argc,
    TSObject * argv[])
{
    Ch * s = new Ch[AsString(argv[0])->Length + 1];
    int sdx;
    FILE * fp;
    Ch b[1024];
    ParseInputStream * pis;
    Ch * t;

    for (sdx = 0; sdx < AsString(argv[0])->Length; sdx++)
        s[sdx] = (Ch) AsString(argv[0])->String[sdx];
    s[sdx] = 0;

    fp = fopen((char *) s, "r");
    if (fp == 0)
    {
        for (t = s; *t != 0; t++)
            if (*t == '\\' || *t == '/')
                goto Failed;

        pis = ReminderStream->GetParseInputStream();
        for (;;)
        {
            strcpy((char *) b, (char *) pis->GetPath());
            strcat((char *) b, (char *) s);

            fp = fopen((char *) b, "r");
            if (fp != 0)
            {
                delete s;
                s = ChStringDup(b);
                break;
            }

            pis = pis->GetPrevious();
            if (pis == 0)
                goto Failed;
        }

    }

    ReminderStream->Include(new StdioInputStream(s, fp));
    return(wld->Unspecified);

Failed:

    delete s;
    wld->ThrowError("Not a valid filename", argv[0]);
    return(wld->Unspecified);
}

void usage()
{
    printf("usage: runrme [OPTION] [DATE]\n");
    printf("    -v               version\n");
    printf("    -b <days>        days before\n");
    printf("    -a <days>        days after\n");
    printf("    -f <filename>    remindme file\n");
    printf("DATE: day, month, and year in any order\n");
    printf("month is jan, feb, ... dec\n\n");
    printf("if no filename is specified, runrme looks for the following:\n");
#ifdef _WIN32_WINNT
    printf("%%USERPROFILE%%\\remindme.rme\n");
    printf("%%USERPROFILE%%\\remindme\\remindme.rme\n");
    printf("My Documents\\remindme.rme\n");
    printf("My Documents\\RemindMe\\remindme.rme\n");
#else
    printf("~/.remindme\n");
    printf("~/remindme.rme\n");
    printf("~/.remindme/remindme.rme\n");
    printf("~/remindme/remindme.rme\n");
    printf("~/Documents/remindme.rme\n");
    printf("~/Documents/RemindMe/remindme.rme\n");
#endif
    exit(1);
}

int main(int argc, char * argv[])
{
    InitRemindMe();

    RMDateType tdt = GetCurrentDate();
    RMDateType cdt = tdt;
    int adx;
    int before = 0;
    int after = 0;
    char * fn = 0;
    FILE * fp;
    unsigned int idx;

    for (adx = 1; adx < argc; adx++)
    {
        if (argv[adx][0] == '-')
        {
            switch (argv[adx][1])
            {
            case 'v':
                printf("runrme 0.1 (%s)\n", __DATE__);
                return(0);
                break;

            case 'b':
                adx += 1;
                before = atoi(argv[adx]);
                break;

            case 'a':
                adx += 1;
                after = atoi(argv[adx]);
                break;

            case 'f':
                adx += 1;
                fn = argv[adx];
                break;

            default:
                usage();
            }
        }
        else
            break;
    }

    if (fn == 0)
    {
        fp = GetDefaultFilename(&fn);
        if (fp == 0)
            usage();
    }
    else
    {
        fp = fopen(fn, "r");
        if (fp == 0)
        {
            printf("fatal error: unable to open %s\n", fn);
            return(1);
        }
    }

    ReminderStream = new RMReminderStream(
            new StdioInputStream(ChStringDup((Ch *) fn), fp));

    for (;adx < argc; adx++)
    {
        int n = atoi(argv[adx]);
        if (n != 0)
        {
            if (n <= 31)
                cdt.DayOfMonth = n;
            else
                cdt.Year = n;
        }
        else
        {
            for (idx = 1; idx <= 12; idx++)
                if (strcasecmp(argv[adx], ShortMonth[idx]) == 0)
                {
                    cdt.Month = idx;
                    break;
                }
        }
    }

    cdt.DayOfWeek = DayOfWeek(cdt.DayOfMonth, cdt.Month, cdt.Year);

    unsigned int rc;
    RMReminderType * rv[1024];

    for (rc = 0; rc < sizeof(rv) / sizeof(RMReminderType *);)
    {
        try
        {
            rv[rc] = ReminderStream->ParseReminder();
            if (rv[rc] == 0)
                break;
            rc += 1;
        }
        catch (RMErrorType)
        {
            if (Error.Source != 0)
                printf("%s(%d): ", Error.Source, Error.LineNumber);
            if (Error.Fatal)
            {
                printf("fatal error: %s\n", Error.Message);
                break;
            }

            printf("error: %s\n", Error.Message);
        }
    }

    unsigned int g = cdt.Gregorian();
    RMDateType sdt(g - before);
    RMDateType edt(g + after);

    for (;;)
    {
        sdt.DayOfWeek = DayOfWeek(sdt.DayOfMonth, sdt.Month, sdt.Year);

        for (idx = 0; idx < rc; idx++)
        {
            if (rv[idx]->Specifier->Match(sdt, 0)) // 1))
            {
                for (adx = 0; adx < rv[idx]->ActionCount; adx++)
                {
                    if (before != 0 || after != 0)
                    {
                        if (sdt == tdt)
                            printf("  ---today---  : ");
                        else
                            printf("%s %2d %s %d: ",
                                    ShortDayOfWeek[sdt.DayOfWeek],
                                    sdt.DayOfMonth, ShortMonth[sdt.Month],
                                    sdt.Year);
                    }

                    rv[idx]->Actions[adx]->Trigger(sdt,
                            rv[idx]->Specifier->GetBaseDate(sdt),
                            rv[idx]->Specifier->MatchDate);
                }
            }
        }

        if (sdt == edt)
            break;
        sdt.Tomorrow();
    }

    fclose(fp);
    return(0);
}
