/*

(Yet Another) Tiny Scheme

*/

#ifndef __TSCM_HPP__
#define __TSCM_HPP__

typedef unsigned short TSCh;
typedef int TSFixnum;

#define Undefined ((TSObject *) 0)

enum TSObjectType
{
    SymbolType,
    NumberType,
    BooleanType,
    CharacterType,
    UnspecifiedType,
    EmptyListType,
    EndOfFileType,
    StringType,
    PairType,
    ProcedureType,
    ContextType,
    BuiltinType,
    SyntaxType,
    VectorType
};

#define HeapP(obj) ((((TSFixnum) (obj)) & 0x3) == 0x0)

class TSWorld;
class TSObject
{
public:

    TSObjectType Type;
    virtual TSObject * Call(TSWorld * wld, int argc, TSObject * argv[]);
    virtual TSObject * Evaluate(TSWorld * wld);
    virtual TSObject * Syntax(TSWorld * wld, TSObject * expr, TSObject * obj);

protected:

    TSObject(TSWorld * wld, TSObjectType ot);
    virtual ~TSObject();

private:

    friend class TSWorld;

    TSObject * Next;
};

class TSString : public TSObject
{
public:

    int Length;
    TSCh * String;

protected:

    virtual ~TSString();

private:

    TSString(TSWorld * wld, int sl, TSCh * s);

    friend class TSWorld;
};

#define StringP(obj) (HeapP((obj)) && ((obj)->Type == StringType))
#define AsString(obj) ((TSString *) (obj))

class TSVector : public TSObject
{
public:

    int Length;
    TSObject ** Vector;

    virtual TSObject * Evaluate(TSWorld * wld);

protected:

    virtual ~TSVector();

private:

    TSVector(TSWorld * wld, int vl, TSObject * obj);

    friend class TSWorld;
};

#define VectorP(obj) (HeapP((obj)) && ((obj)->Type == VectorType))
#define AsVector(obj) ((TSVector *) (obj))

class TSSymbol : public TSObject
{
public:

    TSString * Name;
    TSObject * Value;

    virtual TSObject * Evaluate(TSWorld * wld);
    void SetValue(TSWorld * wld, TSObject * val);

private:

    TSSymbol(TSWorld * wld, TSString * name, TSSymbol * sym);

    friend class TSWorld;

    TSSymbol * Next;
};

#define SymbolP(obj) (HeapP((obj)) && ((obj)->Type == SymbolType))
#define AsSymbol(obj) ((TSSymbol *) (obj))

class TSPair : public TSObject
{
public:

    TSObject * First;
    TSObject * Rest;

    virtual TSObject * Evaluate(TSWorld * wld);

private:

    TSPair(TSWorld * wld, TSObject * first, TSObject * rest);

    friend class TSWorld;
};

#define PairP(obj) (HeapP((obj)) && ((obj)->Type == PairType))
#define AsPair(obj) ((TSPair *) (obj))
#define First(obj) (AsPair(obj)->First)
#define Rest(obj) (AsPair(obj)->Rest)

#define NumberP(obj) ((((TSFixnum) (obj)) & 0x3) == 0x1)
#define AsFixnum(obj) (((TSFixnum) (obj)) >> 2)
#define CharacterP(obj) ((((TSFixnum) (obj)) & 0x3) == 0x2)
#define AsCharacter(obj) (((TSCh) (TSFixnum) (obj)) >> 2)

inline TSObjectType ObjectType(TSObject * obj)
{
    if (NumberP(obj)) return(NumberType);
    if (CharacterP(obj)) return(CharacterType);
    return(obj->Type);
}

typedef TSObject * (*TSBuiltinFn)(TSWorld * wld, int argc, TSObject * argv[]);

struct TSBuiltinInfo
{
    TSBuiltinInfo(char * name, TSBuiltinFn fn, char * sig, char * filename,
            int ln);

    char * Name;
    TSBuiltinFn Builtin;
    char * Signature;
    char * Filename;
    int LineNumber;

    TSBuiltinInfo * Next;
};

static TSBuiltinInfo * BuiltinInfoList = 0;

#define AddAllBuiltins(wld) \
    { \
        for (TSBuiltinInfo * bi = BuiltinInfoList; bi != 0; bi = bi->Next) \
            wld->AddBuiltin(bi); \
    }

#define DefineBuiltin(name, fn, sig) \
    static TSObject * fn ## Builtin (TSWorld * wld, int argc, TSObject * argv[]); \
    static TSBuiltinInfo fn ## Info(name, fn ## Builtin, sig, __FILE__, __LINE__); \
    TSObject * fn ## Builtin

class TSBuiltin : public TSObject
{
public:

    TSBuiltinInfo * Info;

    virtual TSObject * Call(TSWorld * wld, int argc, TSObject * argv[]);

private:

    friend class TSWorld;

    TSBuiltin(TSWorld * wld, TSBuiltinInfo * bi);
};

#define BuiltinP(obj) (HeapP(obj) && ((obj)->Type == BuiltinType))
#define AsBuiltin(obj) ((TSBuiltin *) (obj))

typedef TSObject * (*TSSyntaxFn)(TSWorld * wld, TSObject * expr,
    TSObject * obj);

struct TSSyntaxInfo
{
    TSSyntaxInfo(char * name, TSSyntaxFn fn);

    char * Name;
    TSSyntaxFn Syntax;

    TSSyntaxInfo * Next;
};

static TSSyntaxInfo * SyntaxInfoList = 0;

#define AddAllSyntax(wld) \
    { \
        for (TSSyntaxInfo * si = SyntaxInfoList; si != 0; si = si->Next) \
            wld->AddSyntax(si); \
    }

#define DefineSyntax(name, fn) \
    static TSObject * fn ## Syntax (TSWorld * wld, TSObject * expr, TSObject * obj); \
    static TSSyntaxInfo fn ## Info(name, fn ## Syntax); \
    TSObject * fn ## Syntax

class TSSyntax : public TSObject
{
public:

    char * Name;
    TSSyntaxFn SyntaxFn;

    virtual TSObject * Syntax(TSWorld * wld, TSObject * expr, TSObject * obj);

private:

    friend class TSWorld;

    TSSyntax(TSWorld * wld, char * name, TSSyntaxFn fn);
};

#define SyntaxP(obj) (HeapP(obj) && ((obj)->Type == SyntaxType))
#define AsSyntax(obj) ((TSSyntax *) (obj))

struct TSLineNumber
{
    int LineNumber;
    TSObject * Object;
};

class TSLineNumberMap
{
public:

    TSCh * Filename;
    TSLineNumber * Map;
    int MapLength;

    TSLineNumberMap(TSCh * fn);
    TSLineNumberMap(TSLineNumberMap * lnm);
    ~TSLineNumberMap();
    TSObject * Record(TSObject * obj, int ln);

private:

    int MapSize;
};

class TSContext;
class TSProcedure : public TSObject
{
public:

    TSObject * Name;
    TSObject * Formals;
    short ArgCount;
    short RestFlag;
    TSObject * Body;
    TSContext * Enclosing;
    TSLineNumberMap * LineNumberMap;

    virtual TSObject * Call(TSWorld * wld, int argc, TSObject * argv[]);

protected:

    virtual ~TSProcedure();

private:

    friend class TSWorld;

    TSProcedure(TSWorld * wld, TSObject * name, TSObject * formals,
            TSObject * body, TSLineNumberMap * lnm);
};

#define ProcedureP(obj) (HeapP(obj) && ((obj)->Type == ProcedureType))
#define AsProcedure(obj) ((TSProcedure *) (obj))

class TSInputStream
{
public:

    virtual TSCh GetCh(int * eof) = 0;
    virtual void UnGetCh(TSCh ch) = 0;
    virtual TSCh * GetFilename() = 0;
    virtual int GetLineNumber() = 0;
};

class TSOutputStream
{
public:

    virtual void PutCh(TSCh ch) = 0;
    virtual void PutString(TSCh * s, int sl) = 0;
    virtual void PutString(char * s) = 0;
};

class TSApplication
{
public:

    virtual void FatalError() = 0;
    virtual int Write(TSOutputStream * os, TSObject * obj) = 0;
    virtual int Display(TSOutputStream * os, TSObject * obj) = 0;
};

class TSWorld
{
public:

    TSWorld(TSOutputStream * os, TSApplication * app);
    ~TSWorld();

    void ThrowError(char * es, TSObject * obj);
    void MustBe(TSObjectType ot, TSObject * obj);
    TSObject * Evaluate(TSObject * obj, TSLineNumberMap * lnm);
    TSObject * Evaluate(TSObject * obj) {return(Evaluate(obj, 0));}
    TSObject * Read(TSInputStream * is, TSLineNumberMap * lnm);
    TSObject * Read(TSInputStream * is) {return(Read(is, 0));}
    void Write(TSOutputStream * os, TSObject * obj);
    void Display(TSOutputStream * os, TSObject * obj);
    TSObject * InternSymbol(int sl, TSCh * s);
    TSObject * InternSymbol(char * s);
    TSObject * MakeString(int sl, TSCh * s);
    TSObject * MakeVector(int vl, TSObject * obj);
    TSObject * MakePair(TSObject * first, TSObject * rest);
    TSObject * MakeNumber(TSFixnum n);
    TSObject * MakeCharacter(TSCh ch);
    TSObject * MakeProcedure(TSObject * name, TSObject * formals,
            TSObject * body, TSLineNumberMap * lnm);
    TSObject * MakeProcedure(TSObject * name, TSObject * formals,
            TSObject * body);
    void AddBuiltin(TSBuiltinInfo * bi);
    void AddSyntax(TSSyntaxInfo * si);

    TSObject * False;
    TSObject * True;
    TSObject * Unspecified;
    TSObject * EmptyList;
    TSObject * EndOfFile;

    TSObject * SymbolQuote;
    TSObject * SymbolEqualArrow;
    TSObject * SymbolElse;
    TSObject * SymbolDefine;

    TSContext * Stack;

    TSOutputStream * DefaultOutput;

    TSApplication * Application;

private:

    TSObject * ReadInternal(TSInputStream * is, TSLineNumberMap * lnm, int eaf,
            int rlf);
    TSObject * ReadString(TSInputStream * is);
    TSObject * ReadSymbolOrNumber(TSInputStream * is, TSCh ch);
    TSObject * ReadList(TSInputStream * is, TSLineNumberMap * lnm);

    friend class TSObject;

    TSObject * Objects;
    TSSymbol * Symbols[67];
};

TSObject * Evaluate(TSWorld * wld, TSObject * obj);

#endif // __TSCM_HPP__
